/**
 * Egyszer osztlyhierarchia az ctor/dtor feladatainak bemutatshoz
 * A pointerek tlltst csak kzvetve, az f() virtulis fggvnyen
 * keresztl tudjuk megfigyelni, amit mindig a g() fggvny hv az alaposztlybl.
 * A g() fggvnyt viszont minden konstruktorbl s destruktorbl meghvunk.
 * gy Attl fggen, hogy az f() virt. fggvny pointere hova mutat,
 * ms s ms f() vltozat futnak le.
 */

 /**
 * HA a DUMP makr definilt, akkor a g() fggvny az alaposztly
 * memriakpt is kirja hexadecimlisan.
 * FELTTELEZZK, hogy a memriakp elejn a virt. tbla mutatja ll.
 * Ezt kvetik az adattagok. Az egyszersg kedvrt 3 db pointer az
 * adattag amit 1,2,3 rtkkel tltnk fel.
 */
#define DUMP

#include <iostream>

using std::cout;
using std::cerr;
using std::cin;
using std::endl;

/**
 Konstruktor:
 * rklsi lnc vgn hvja a virtulis alaposztlyok konstruktorait.
 * Hvja a kzvetlen, nem virtulis alaposztlyok konstruktorait.
 * Ltrehozza a sajt rszt:
   - belltja a virtulis alaposztly mutatit
   - belltja a virtulis fggvnyek mutatit
   - hvja a tartalmazott objektumok konstruktorait
   - vgrehajtja a programozott trzset
 */


/**
 Destruktor:
 * Megsznteti a sajt rszt:
   - vgrehajtja a programozott trzset
   - hvja a tartalmazott objektumok destruktorait
   - belltja a virtulis fggvnyek mutatit
   - belltja a virtulis alaposztly mutatit
 * Hvja a kzvetlen, nem virtulis alaposztlyok destruktorait.
 * rklsi lnc vgn hvja a virtulis alaposztlyok destruktorait.
 */


/// Virtulis alaposztly
struct VirtAlap {
    int* adat[3];   /// dump miatt 3 db int*
    VirtAlap() {
        for (int i = 0; i < 3; i++) adat[i] = reinterpret_cast<int*>(i+1);
        cout << "Oroklesi lanc vegen virt. alaposztaly konstruktor" << endl;
        g("VirtAlap()");
    }
    ~VirtAlap() {
        cout << "Oroklesi lanc vegen alaposztaly destruktor" << endl;
        g("~VirtAlap()");
    }
    void g(const char *honnan) {
        cout << "  " << honnan << "-bol hivjuk f()-et:";
        f();
        dump();
    }
    virtual void f() {
        cout << " VirtAlap::f() vagyok" << endl;
    }
    /// p0 cmtl l darab int*-ot r ki
    void dump(void*p0, int l) {
        int **p = reinterpret_cast<int**>(p0);
        for (int i = 0; i < l; i++)
            cout << *p++ << " ";
        cout << endl;
    }

    void dump() {
#ifdef DUMP
        cout << "    Alaposztaly memoriakepe: ";
        dump(this, sizeof(VirtAlap)/sizeof(int*));
        int* p = *reinterpret_cast<int**>(this);
        cout << "    Virt.fv. tabla tartalma: ";
        dump(p, 2);
#endif
    }
};

/// Kzvetlen nem virt alaposztly
struct KozvNemVirtAlap : virtual  VirtAlap {
    KozvNemVirtAlap()  {
        cout << "Kozvetlen nem virt. alaposztaly konstruktor" << endl;
        g("KozvNemVirtAlap()");
    }
    ~KozvNemVirtAlap() {
        cout << "Kozvetlen nem virt. alaposztaly destruktor" << endl;
        g("~KozvNemVirtAlap()");
    }
    void f() {
        cout << " KozvNemVirtAlap::f() vagyok" << endl;
    }
};

/// Adattag
struct Adat {
    Adat()  { cout << "Tartalmazott obj. konstruktor" << endl; }
    ~Adat() { cout << "Tartalmazott obj. destruktor" << endl; }
};

/// Az OSZTALY
struct A : KozvNemVirtAlap {
    Adat adat;
    A()  {
        cout << "Konstruktor programozott torzse" << endl;
        g("A()");
    }
    ~A() {
        cout << "Destruktor programozott torzse" << endl;
        g("~A()");
    }
    void f() {
        cout << " A::f() vagyok" << endl;
    }
};

int main() {
    cout << "struct VirtAlap;\nstruct KozvNemVirtAlap : virtual VirtAlap;\n"
         << "struct A : KozvNemVirtAlap;\n"
         << "\nA* p = new A;\n\n";
    A* p = new A;
    cin.ignore(100, '\n');
    cout << "delete p;\n\n";
    delete p;
    return 0;
}

