/*
 * \file polimorf.cpp
 *
 * Tbbszrs rkls s a polimorf viselkeds,
 * valamint a dynamic_cast memutatsa.
 *
 * A dynamic_cast-hoz fontos, hogy a fordtskor legyen
 * engedlyezve a Run-Time Type Info (RTTI) generlsa
 * (Vis.stud: /GR opci)
 * GCC-nl automatikus
 *
 */

#include <iostream>
#include <stdexcept>

using namespace std;

/// A oszty.
/// (struct: mindene publikus)
struct A {
    void f()        { cout << " A::f()\n"; }
    void f(int)     { cout << " A::f(int)\n";  }
    virtual ~A() {} // csak ezrt virt., hogy legyen RTTI inf
};

/// B oszty.
/// (struct: mindene publikus)
struct B {
    void f(double)  { cout << " B::f(double)\n"; };
};


/// C oszty.
/// (struct: mindene publikus)
struct C {
    void f(int*)  { cout << " C::f(int*)\n"; };
};

/// AB oszty.
class AB :public A, public B {
public:
    // Tbbszrs rklsnl az azonos nev tagfggvnyekrl a fordt felttelezi,
    // hogy teljesen eltr viselkedst valstanak meg, ezrt egyik elrst sem
    // engedi automatikusan, hiba publikus az rkls.
    using A::f;     // gy elrhetk az A-beli f()-ek
    using B::f;     // gy elrhet a B-beli f()is
    void f() {
        cout << " AB::f()\n";
        A::f();
        f(1);       // A::f(int)
        f(1.2);     // B::f(double)
    }
};

/// Kiprobljuk...
int main() {
    AB ab;

    cout << "Fuggvenyek [ab.A::f(), ab.f(int) ab.f(doube) ab.f()]" << endl;
    ab.A::f();
    ab.f(5);        // A::f(int)
    ab.f(6.3);      // B::f(double)
    ab.f();         // AB::f()

    A *Ap = &ab;    // az upcast automaiikus
    B *Bp = &ab;    // az upcast automaiikus

    cout << endl << "Pointerek konvertalasa:" << endl;
    AB *ABp = dynamic_cast<AB*>(Ap);    // biztonsgos down cast
    if (ABp != 0) {
        cout << "Sikeres volt a down cast" << endl;
        ABp->f();
    } else {
        cout << "A down cast nem sikerult" << endl;
    }

    C *Cp = dynamic_cast<C*>(Ap);        // ez nem fog menni
    if (Cp == 0) cout << "AB* nem konvertalhato C*-ra" << endl;

    cout << endl << "Referencia konvertalasa:" << endl;
    // Referencia konvertlsnl a hibajelzs kivetellel jn
    try {
        AB& ab2 = dynamic_cast<AB&>(*Ap);
        cout << "Sikeres volt a down cast referenciara is:" << endl;
        ab2.f();
        cout << "Kiserlet C referenciara valo konvertziora:" << endl;
        C& c2 = dynamic_cast<C&>(*Ap);
    } catch (exception& e) {
        cout << e.what() << endl;
    }
    return 0;
}
